`include "control.h"
`include "memory.h"
`include "dbgflags.h"

module CacheControl (
        DStrobe, DRW, DReady,
        Match, Valid,
        Write,
        CacheDataSelect,
        DDataSelect,
	Hit,   //signals a  cache Hit
	Miss, //signals a miss 
        MDataOE, DDataOE,
        MStrobe, MRW, Reset, Clk );

input           DStrobe, DRW;
output          DReady;
input           Match, Valid;
output          Write;
output          CacheDataSelect;
output          DDataSelect;
output          Hit;	    //NEW--Added for set-associative cache
output          Miss;	    //NEW--Added for set-associative cache
output          MDataOE, DDataOE;
output          MStrobe, MRW;
input           Reset;
input           Clk;


reg             WSCLoad;
reg    [15:0]   WSCLoadVal;
wire            WSCSig;
WaitStateCtr    WaitStateCtr (WSCLoad, WSCLoadVal, WSCSig, Clk);

reg             DReadyEnable;
reg             MStrobe;
reg             MRW;
reg             MDataOE;
reg             Write;
reg             Ready;
reg             CacheDataSelect;
reg             DDataSelect;
reg             Hit;
reg             Miss;
reg             DDataOE;

wire DReady = (DReadyEnable && Match && Valid && DRW) || Ready;

/*
always @ (DReadyEnable or Match or Valid or Ready)
   if (DReadyEnable && Match && Valid && DRW)
      $display($time, " DReady due to Read Cache Hit");
   else if (Ready)
      $display($time, " DReady due to Memory Ready");
   else
      $display($time, " DReady value %b due to (%b,%b,%b,%b)",
                DReady, DReadyEnable, Match, Valid, Ready);
*/

reg     [3:0]   State;
reg     [3:0]   NextState;


initial begin
   State     = `STATE_IDLE;
   NextState = `STATE_IDLE;
end

always @ (posedge Clk) begin
   State = NextState;
   UpdateSignals(State);
   if (Reset)
      NextState = `STATE_IDLE;
   else begin
      case (State)
         `STATE_IDLE: begin
                        if (`dbg) $display(" ctrl> IDLE", $time);
                        if (DStrobe && DRW)
                           NextState = `STATE_READ;
                        else if (DStrobe && !DRW)
                           NextState = `STATE_WRITE;
                        else
                           NextState = `STATE_IDLE;
                     end

         `STATE_READ: begin
                        if (`dbg) $display(" ctrl> READ", $time);
                        if (Match && Valid) begin    // read hit
                           NextState = `STATE_IDLE;
			    Hit = 1;  //Kludge since there is no separate READ-HIT state
                           if (`vbs) $display(" ctrl> READ HIT", $time);
                           Main.Driver.nreadhits = Main.Driver.nreadhits + 1;
                        end
                        else                         // read miss
                           NextState = `STATE_READMISS;
                     end

         `STATE_READMISS: begin
                        if (`vbs) $display(" ctrl> READMISS", $time);
                        Main.Driver.nreadmiss = Main.Driver.nreadmiss + 1;
                        WSCLoadVal = `READ_WAITCYCLES-1;
                        NextState  = `STATE_READMEM;
                     end

         `STATE_READMEM: begin
                        if (`dbg) $display(" ctrl> READMEM", $time);
                        if (WSCSig)
                           NextState = `STATE_READDATA;
                        else
                           NextState = `STATE_READMEM;
                     end

         `STATE_READDATA: begin
                        if (`dbg) $display(" ctrl> READDATA", $time);
                        NextState = `STATE_IDLE;
                     end

         `STATE_WRITE: begin
                        if (`dbg) $display(" ctrl> WRITE", $time);
                        if (Match && Valid) begin
                           NextState = `STATE_WRITEHIT;
                        end
                        else
                           NextState = `STATE_WRITEMISS;
                     end

         `STATE_WRITEHIT: begin
                        if (`vbs) $display(" ctrl> WRITEHIT", $time);
                        Main.Driver.nwritehits = Main.Driver.nwritehits + 1;
                        WSCLoadVal = `WRITE_WAITCYCLES-2;
                        NextState  = `STATE_WRITEMEM;
                     end

         `STATE_WRITEMISS: begin
                     if (`vbs) $display(" ctrl> WRITEMISS ", $time);
                        Main.Driver.nwritemiss = Main.Driver.nwritemiss + 1;
                        WSCLoadVal = `WRITE_WAITCYCLES-2;
                        NextState  = `STATE_WRITEMEM;
                     end

         `STATE_WRITEMEM: begin
                        if (`dbg) $display(" ctrl> WRITEMEM", $time);
                        if (WSCSig)
                           NextState = `STATE_WRITEDATA;
                        else
                           NextState = `STATE_WRITEMEM;
                     end

         `STATE_WRITEDATA: begin
                        if (`dbg) $display(" ctrl> WRITEDATA", $time);
                        NextState = `STATE_IDLE;
                     end
      endcase
   end
end

task OutputVec;
input   [11:0]   vector;
begin
   Miss                 = vector[11];   //NEW--Added for 4-way cache 
   Hit                  = vector[10];  //NEW--Added for 4-way cache
   WSCLoad              = vector[9];
   DReadyEnable         = vector[8];
   Ready                = vector[7];     // signal driver
   Write                = vector[6];     // to cache rams
   MStrobe              = vector[5];     // memory strobe
   MRW                  = vector[4];
   CacheDataSelect      = vector[3];
   DDataSelect          = vector[2];
   DDataOE              = vector[1];
   MDataOE              = vector[0];

end
endtask

task UpdateSignals;
input [3:0] state;
   case (state)
      `STATE_IDLE:      OutputVec(12'b000000000000);
      `STATE_READ:      OutputVec(12'b000100000010); //Hit also set for Read Hit only
      `STATE_READMISS:  OutputVec(12'b101000110010);
      `STATE_READMEM:   OutputVec(12'b000000010010);
      `STATE_READDATA:  OutputVec(12'b000011011110);
      `STATE_WRITE:     OutputVec(12'b000100000000);
      `STATE_WRITEHIT:  OutputVec(12'b011001101101);
      `STATE_WRITEMISS: OutputVec(12'b101001100001);
      `STATE_WRITEMEM:  OutputVec(12'b000000000001);
      `STATE_WRITEDATA: OutputVec(12'b000010001101);
   endcase
endtask

endmodule /* Control */



module WaitStateCtr(Load, LoadValue, Sig, Clk);
input           Load;
input   [15:0]  LoadValue;
output          Sig;
input           Clk;

reg     [15:0]  Count;
wire            Sig = Count == 16'b0;

always @(posedge Clk) begin
   if (Load)
      Count = LoadValue;
   else
      Count = Count - 16'b1;
   if (`dbg) $display($time, " WaitState    %d", Count);
end

endmodule /* WaitStateCtr */
